/*
Data Purity Investigation

STEP 1
Run this script within the context of the database you want to check (use [dbname];)
Once it has finished, which may take several hours if there are a lot of data purity errors, run the commented out SQL under step 2
to view the results, or use step 3 to generate UPDATE statements (be careful - read the article first).
Finally, run the commented out code under step 4 to tidy up.
*/ 
SET NOCOUNT ON; 
SET ANSI_WARNINGS OFF; 

DECLARE 
@db sysname , 
@ver varchar ( 20 ) , 
@sql varchar ( max ) , 
@record varchar ( 255 ) 

SELECT @db = db_name ( ) 

SELECT @ver = convert ( varchar ( 20 ) , serverproperty ( 'ProductVersion' ) ) 

SELECT @sql = 'DBCC CHECKDB (''' + @db + ''') WITH DATA_PURITY, NO_INFOMSGS, TABLERESULTS' 

-- Create temporary tables. These will need to be dropped manually once the results of the script have analysed

CREATE TABLE tmp_dbcc_results 
( 
    [ObjectId] int NULL , 
    [Column] sysname NULL , 
    [File] int NULL , 
    [Page] int NULL , 
    [Slot] int NULL 
) 

CREATE TABLE tmp_page_results 
( 
    [ParentObject] sysname NULL , 
    [Object] sysname NULL , 
    [Field] sysname NULL , 
    [Value] sysname NULL 
) 

CREATE NONCLUSTERED INDEX idx_slot 
ON tmp_page_results ( [Object] , [Field] ) 
INCLUDE ( [Value] ) 

CREATE TABLE tmp_final_results 
( 
    [Id] int IDENTITY ( 1 , 1 ) PRIMARY KEY CLUSTERED , 
    [Table] sysname NULL , 
    [Key] varchar ( 1000 ) NULL , 
    [Column] sysname NULL , 
    [Type] varchar ( 50 ) NULL , 
    [Value] sql_variant NULL 
) 

CREATE NONCLUSTERED INDEX idx_val 
ON tmp_final_results ( [Value] , [Table] ) 

CREATE TABLE tmp_primary_keys 
( 
    [Table] sysname , [Column] sysname , [Type] varchar ( 50 ) 
) 

-- Perform the data purity check. The output is version dependent, hence the condition code
IF @ver LIKE '9%' OR @ver LIKE '10%' 
BEGIN 
    CREATE TABLE tmp_dbcc_results_to_2008_r2 
    ( 
        [Error] int NULL , 
        [Level] int NULL , 
        [State] int NULL , 
        [MessageText] nvarchar ( 2048 ) NULL , 
        [RepairLevel] nvarchar ( 100 ) NULL , 
        [Status] int NULL , 
        [DbId] int NULL , 
        [ObjectId] int NULL , 
        [IndexId] int NULL , 
        [PartitionId] bigint NULL , 
        [AllocUnitId] bigint NULL , 
        [File] int NULL , 
        [Page] int NULL , 
        [Slot] int NULL , 
        [RefFile] int NULL , 
        [RefPage] int NULL , 
        [RefSlot] int NULL , 
        [Allocation] int NULL 
    ) 

    INSERT tmp_dbcc_results_to_2008_r2 

    EXEC ( @sql ) 

    INSERT tmp_dbcc_results 
    SELECT 
    [ObjectId] , 
    substring ( MessageText , charindex ( '). Column "' , MessageText ) + 11 , 
    charindex ( '"' , MessageText , charindex ( '). Column "' , MessageText ) + 11 ) - 
    charindex ( '). Column "' , MessageText ) - 11 ) , 
    [File] , 
    [Page] , 
    [Slot] 
    FROM tmp_dbcc_results_to_2008_r2 WHERE [Error] = 2570 

    DROP TABLE tmp_dbcc_results_to_2008_r2 
END 
ELSE IF @ver LIKE '11%' 
BEGIN 
    CREATE TABLE tmp_dbcc_results_from_2012 
    ( 
        [Error] int NULL , 
        [Level] int NULL , 
        [State] int NULL , 
        [MessageText] nvarchar ( 2048 ) NULL , 
        [RepairLevel] nvarchar ( 100 ) NULL , 
        [Status] int NULL , 
        [DbId] int NULL , 
        [DbFragId] int NULL , 
        [ObjectId] int NULL , 
        [IndexId] int NULL , 
        [PartitionId] bigint NULL , 
        [AllocUnitId] bigint NULL , 
        [RidDbId] int NULL , 
        [RidPruId] int NULL , 
        [File] int NULL , 
        [Page] int NULL , 
        [Slot] int NULL , 
        [RefDbId] int NULL , 
        [RefPruId] int NULL , 
        [RefFile] int NULL , 
        [RefPage] int NULL , 
        [RefSlot] int NULL , 
        [Allocation] int NULL 
    ) 

    INSERT tmp_dbcc_results_from_2012 

    EXEC ( @sql ) 

    INSERT tmp_dbcc_results 
    SELECT 
    [ObjectId] , 
    substring ( MessageText , charindex ( '). Column "' , MessageText ) + 11 , 
    charindex ( '"' , MessageText , charindex ( '). Column "' , MessageText ) + 11 ) - 
    charindex ( '). Column "' , MessageText ) - 11 ) , 
    [File] , 
    [Page] , 
    [Slot] 
    FROM tmp_dbcc_results_from_2012 WHERE [Error] = 2570 

    DROP TABLE tmp_dbcc_results_from_2012 
END 

-- Run DBCC PAGE for each error.
-- CHECKDB only generates one error per row, even if there are multiple columns with data purity issues.
-- This code checks the DBCC PAGE output to ensure these missing errors are captured.
DECLARE 
@ObjectId int , 
@Column nvarchar ( 255 ) , 
@File int , 
@Page int , 
@Slot int 

DECLARE c CURSOR FOR 
SELECT 
[ObjectId] , 
[Column] , 
[File] , 
[Page] , 
[Slot] 
FROM tmp_dbcc_results 

OPEN c 

FETCH NEXT FROM c INTO @ObjectId , @Column , @File , @Page , @Slot 

WHILE @@FETCH_STATUS = 0 
BEGIN 
    SELECT @sql = 'DBCC PAGE (' + @db + ', ' + convert ( varchar ( 10 ) , @File ) 
    + ', ' + convert ( varchar ( 20 ) , @Page ) + ', 3) WITH TABLERESULTS' 

    INSERT tmp_page_results 

    EXEC ( @sql ) 

    IF NOT EXISTS 
    ( 
        SELECT 1 FROM tmp_primary_keys WHERE [Table] = object_name ( @ObjectId ) 
    ) 
    BEGIN 
        INSERT tmp_primary_keys 
        ( 
            [Table] , [Column] , [Type] 
        ) 
        SELECT 
        object_name ( @ObjectId ) , 
        cu.COLUMN_NAME , 
        t.name 
        FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE cu 
        JOIN sys.columns c ON c.object_id = @Objec 
        AND c.name = cu.COLUMN_NAME 
        JOIN sys.types t ON c.user_type_id = t.user_type_id 
        JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc ON tc.TABLE_NAME = object_name ( @ObjectId ) 
        AND tc.CONSTRAINT_NAME = cu.CONSTRAINT_NAME 
        AND tc.CONSTRAINT_TYPE = 'PRIMARY KEY' 

    END 

    SELECT @record = 'Slot ' + convert ( varchar ( 10 ) , @Slot ) + ' %' 

    INSERT tmp_final_results 
    ( 
        [Table] , [Key] , [Column] , [Type] , [Value] 
    ) 
    SELECT 
    object_name ( @ObjectId ) , 
    ( 
        SELECT left ( x.pk , len ( x.pk ) - 4 ) FROM 
        ( 
            SELECT '[' + pk. [Column] + '] = ' + 
            CASE 
                WHEN pk. [Type] LIKE '%char%' OR pk. [Type] LIKE '%date%' THEN '''' 
                ELSE '' 
            END + pr.Value + 
            CASE 
                WHEN pk. [Type] LIKE '%char%' OR pk. [Type] LIKE '%date%' THEN '''' 
                ELSE '' 
            END + ' and ' AS [text()] FROM tmp_primary_keys pk 
            JOIN tmp_page_results pr ON pr. [Object] LIKE @record 
            AND pr. [Field] = pk. [Column] 
            WHERE pk. [Table] = object_name ( @ObjectId ) FOR xml path ( '' ) 
        ) 
        x ( pk ) 
    ) , 
    ic.Field , 
    ict.name , 
    NULL 
    FROM tmp_page_results ic 
    JOIN sys.columns icc ON icc.object_id = @ObjectId 
    AND icc.name = ic. [Field] 
    JOIN sys.types ict ON icc.user_type_id = ict.user_type_id 
    WHERE ic. [Object] LIKE @record 
    AND 
    ( ic. [VALUE] = 'INVALID COLUMN VALUE' 
        OR ic. [Field] = @Column ) 

    SELECT @sql = 
    ( 
        SELECT 'update tmp_final_results set [Value] = (select [' + [Column] + '] 
        from [' + [Table] + '] where ' + [Key] + ') where [Id] = ' + convert ( varchar ( 10 ) , [Id] ) + ';' 
        FROM tmp_final_results 
        WHERE [Table] = object_name ( @ObjectId ) 
        AND [Value] IS NULL FOR xml path ( '' ) 
    ) 

    EXEC ( @sql ) 

    TRUNCATE TABLE tmp_page_results 

    FETCH NEXT FROM c INTO @ObjectId , @Column , @File , @Page , @Slot 
END 

CLOSE c 

DEALLOCATE c 

/*
-- STEP 2
-- Run this to view the results.
-- The conversion to varchar is necessary in some cases to prevent arithmetic overflow errors

SELECT [Table], [Key], [Column], [Type], CONVERT(varchar(255), [Value]) AS [Value]
FROM tmp_final_results
*/ 

/*
-- STEP 3
-- Run this to generate the UPDATE statements that will fix the data purity issues
-- IMPORTANT - READ ACCOMPANYING ARTICLE FIRST - DO NOT RUN THE GENERATED UPDATE STATEMENTS UNLESS 100% SURE THEY ARE CORRECT

SELECT 'UPDATE [' + [Table] + '] SET [' + [Column] + '] = ' + CASE
WHEN [Type] LIKE '%char%'
OR [Type] LIKE '%date%' THEN ''''
ELSE ''
END + CONVERT(varchar(255), [Value]) + CASE
WHEN [Type] LIKE '%char%'
OR [Type] LIKE '%date%' THEN ''''
ELSE ''
END + ' WHERE ' + [Key] + '; '
FROM   tmp_final_results
*/ 

/*
-- STEP 4
-- Finally run this bit to drop all temporary tables

DROP TABLE tmp_primary_keys
DROP TABLE tmp_final_results
DROP TABLE tmp_page_results
DROP TABLE tmp_dbcc_results
*/