/***********************************************************************************************************************

SP for determining jobs that have run for more than a set period of time
 
SQL 2005 or greater (tested on 2005, 2008, 2008R2, 2012)


VERSION HISTORY:
Version 1.0
24 Jun 2013

Version 1.0.1
27 Jun 2013
Included runtime in email
Included server uptime in email

Gregory Ferdinandsen
greg@ferdinandsen.com
gregory.ferdinandsen@recall.com

Parameters:
@MaxRunTime = The run time on which the alert has been sent, value in minutes
@EMailAlertAddress = The email alert address
@EMailCCAddress = An alert CC address
@ExcludeJob1 = #1 Job to exclude from alerting
@ExcludeJob2 = #2 Job to exclude from alerting
@ExcludeJob3 = #3 Job to exclude from alerting
@ExcludeJob4 = #4 Job to exclude from alerting
@ExcludeJob5 = #5 Job to exclude from alerting

sp_help_job http://msdn.microsoft.com/en-us/library/ms186722.aspx

examples:
Simple Execution (All default parameters; i.e. no excluded databases, default thresholds, etc):
exec DBMonitor..sp_Long_Running_Jobs

More Complex Exections:
exec DBMonitor..sp_Long_Running_Jobs @MaxRunTime = 720

exec DBMonitor..sp_Long_Running_Jobs
@MaxRunTime = 1440,
@EMailAlertAddress = 'a@b.com',
@EMailCCAddress = 'a@b.org',
@ExcludeJob1 = 'collection_set_2_collection',
@ExcludeJob2 = 'collection_set_3_collection',
@ExcludeJob3 = null,
@ExcludeJob4 = null,
@ExcludeJob5 = null

collection_set_2_collection and collection_set_3_collection are automatically excluded as they are in
a running state since the server started, these two jobs are for the MDW

720 minutes = 12 hours, 1440 minutes = 24 hours
@LogStats = 'true'
***********************************************************************************************************************/ 

CREATE PROCEDURE [dbo].[sp_LongRunning_Jobs] 
( 
    @MaxRunTime smallint = 1440 , 
    @EMailAlertAddress varchar ( 256 ) = 'a@b.com' , 
    @EMailCCAddress varchar ( 256 ) = '' , 
    @ExcludeJob1 varchar ( 1024 ) = NULL , 
    @ExcludeJob2 varchar ( 1024 ) = NULL , 
    @ExcludeJob3 varchar ( 1024 ) = NULL , 
    @ExcludeJob4 varchar ( 1024 ) = NULL , 
    @ExcludeJob5 varchar ( 1024 ) = NULL 
) 
AS 
DECLARE @n AS int 
DECLARE @ErrorText AS varchar ( 256 ) 
DECLARE @RunningJobs AS smallint 
DECLARE @LongRunningJobs AS smallint 
DECLARE @MaxTimeHours AS smallint 
DECLARE @Subject AS varchar ( 128 ) 
DECLARE @msg AS varchar ( 4096 ) 
DECLARE @SPVersion AS sql_variant 
DECLARE @SPName AS varchar ( 128 ) 
DECLARE @SPModify AS datetime 
DECLARE @SPCreate AS datetime 
DECLARE @StartTime AS datetime 
DECLARE @UpTimeSec AS bigint 
DECLARE @UpTimeFormatted AS varchar ( 128 ) 

SET NOCOUNT ON 
SET concat_null_yields_null OFF 

--Verify Parameters
IF @MaxRunTime % 60 <> 0 
BEGIN 
    SET @ErrorText = 'The value for @MaxRunTime must be divisible by 60.' 
    RAISERROR ( @ErrorText , 16 , 1 ) 
    RETURN 
END 

--Determine SP information
SET @SPName = object_name ( @@procid ) 
SELECT @SPCreate = create_date FROM sys.objects WHERE name = @SPName 

SELECT @SPModify = modify_date FROM sys.objects WHERE name = @SPName 

SELECT @SPVersion = value FROM sys.extended_properties WHERE name = 'LongRunning--Version' 

--Determine Server Uptime

SELECT @StartTime = login_time FROM master.sys.dm_exec_sessions WHERE session_id = 1 

SET @UpTimeSec = datediff ( ss , @StartTime , getdate ( ) ) 
SET @UpTimeFormatted = cast ( ( @UpTimeSec / 86400 ) AS varchar ( 8 ) ) + ' day(s) ' + 
cast ( ( @UpTimeSec % 86400 / 3600 ) AS varchar ( 32 ) ) + ' hour(s) ' + 
cast ( ( @UpTimeSec % 3600 / 60 ) AS varchar ( 32 ) ) + ' min(s) ' + 
cast ( ( @UpTimeSec % 60 ) AS varchar ( 32 ) ) + ' sec(s)' 

--Set all variables in the message
SET @msg = 'SP Name: ' + @SPName + char ( 13 ) + 
'SP Version: ' + cast ( @SPVersion AS varchar ( 16 ) ) + char ( 13 ) + 
'SP Created: ' + cast ( @SPCreate AS varchar ( 32 ) ) + char ( 13 ) + 
'SP Modified: ' + cast ( @SPModify AS varchar ( 32 ) ) + char ( 13 ) + char ( 13 ) + 
'Server Uptime: ' + cast ( @UpTimeFormatted AS varchar ) + char ( 13 ) + char ( 13 ) + 
'----------------------------------------------------------------' + char ( 13 ) + char ( 13 ) 

IF EXISTS 
( 
    SELECT * FROM tempdb.sys.objects 
    WHERE name LIKE '##Jobs%' 
    AND type IN ( 'U' ) 
) 
BEGIN 
    DROP TABLE ##Jobs 
END 

--Populate Temp Table
SELECT 
JobName = sj.name , 
StartDate = sja.start_execution_date , 
EndDate = sja.stop_execution_date , 
status = 
CASE 
    WHEN isnull ( sjh.run_status , - 1 ) = - 1 AND sja.start_execution_date IS NULL AND sja.stop_execution_date IS NULL THEN 'Idle' 
    WHEN isnull ( sjh.run_status , - 1 ) = - 1 AND sja.start_execution_date IS NOT NULL AND sja.stop_execution_date IS NULL THEN 'Running' 
    WHEN isnull ( sjh.run_status , - 1 ) = 0 THEN 'Failed' 
    WHEN isnull ( sjh.run_status , - 1 ) = 1 THEN 'Succeeded' 
    WHEN isnull ( sjh.run_status , - 1 ) = 2 THEN 'Retry' 
    WHEN isnull ( sjh.run_status , - 1 ) = 3 THEN 'Canceled' 
END 
, 
RunTime = datediff ( mi , sja.start_execution_date , getdate ( ) ) 
INTO ##Jobs 
FROM msdb.dbo.sysjobs sj 
JOIN msdb.dbo.sysjobactivity sja ON sj.job_id = sja.job_id 
JOIN 
( 
    SELECT MaxSessionid = max ( session_id ) FROM msdb.dbo.syssessions 
) 
ss ON ss.MaxSessionid = sja.session_id 
LEFT JOIN msdb.dbo.sysjobhistory sjh ON sjh.instance_id = sja.job_history_id 

--Delete MDW jobs from Temp Table, delete jobs that are not running

DELETE FROM ##Jobs WHERE JobName IN 
( 'collection_set_2_collection' , 'collection_set_3_collection' , 
    @ExcludeJob1 , @ExcludeJob2 , @ExcludeJob3 , @ExcludeJob4 , @ExcludeJob5 ) 
DELETE FROM ##Jobs WHERE status <> 'Running' 
DELETE FROM ##Jobs WHERE status IS NULL 

--Delete Jobs running shorter than selected interval
DELETE FROM ##Jobs WHERE RunTime <= @MaxRunTime 
SELECT @LongRunningJobs = count ( * ) FROM ##Jobs 

--Exit SP if @LongRunningJobs is zero

IF @LongRunningJobs = 0 RETURN 

SET @MaxTimeHours = @MaxRunTime / 60 

--Else @LongRunningJobs is greater than or equal to one (inferred)
SET @Subject = cast ( @LongRunningJobs AS varchar ) + ' Long Running Job(s) on ' + @@servername 
SET @msg = @msg + 'The following jobs have been running for more than ' + ( cast ( @MaxTimeHours AS varchar ) ) + ' hour(s):' + char ( 13 ) + char ( 13 ) 
SELECT @msg = @msg + JobName + ' ---> Running For ' + cast ( RunTime / 60 AS varchar ) + ' Hours and ' + cast ( Runtime % 60 AS varchar ) + 
' minutes(s) --> Since ' + cast ( StartDate AS varchar ) + char ( 13 ) FROM ##Jobs 

--Email the SQL Group a list of jobs running over nn hours

EXEC msdb..sp_send_dbmail 
@recipients = @EMailAlertAddress , 
@copy_recipients = @EMailCCAddress , 
@body = @msg , 
@subject = @Subject 

--Drop the Temp Table in preparation for the next execution of the SP
IF EXISTS 
( 
    SELECT * FROM tempdb.sys.objects 
    WHERE name LIKE '##RunningJobs%' 
    AND type IN ( 'U' ) 
) 
BEGIN 
    DROP TABLE ##RunningJobs 
END 
GO 

EXEC sys.sp_addextendedproperty @name = N'LongRunning--Version' , @value = N'1.0.1' , @level0type = N'SCHEMA' , @level0name = N'dbo' , @level1type = N'PROCEDURE' , @level1name = N'sp_LongRunning_Jobs' 
GO 

EXEC sys.sp_addextendedproperty @name = N'LongRunning--Author' , @value = N'Gregory Ferdinandsen' , @level0type = N'SCHEMA' , @level0name = N'dbo' , @level1type = N'PROCEDURE' , @level1name = N'sp_LongRunning_Jobs' 
GO